@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2010 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2011 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -69,7 +69,9 @@ extern inline int wcscasecmp(const wchar_t* s1, const wchar_t* s2);
 #define WRAPPER_THREAD_MAIN     1
 #define WRAPPER_THREAD_SRVMAIN  2
 #define WRAPPER_THREAD_TIMER    3
-#define WRAPPER_THREAD_COUNT    4
+#define WRAPPER_THREAD_JAVAIO   4
+#define WRAPPER_THREAD_STARTUP  5
+#define WRAPPER_THREAD_COUNT    6
 
 #define MAX_LOG_SIZE 4096
 
@@ -141,6 +143,26 @@ extern void outOfMemoryQueued(const TCHAR *context, int id);
 
 extern void outOfMemory(const TCHAR *context, int id);
 
+#ifdef _DEBUG
+/**
+ * Used to dump memory directly to the log file in both HEX and readable format.
+ *  Useful in debugging applications to track down memory overflows etc.
+ *
+ * @param label A label that will be prepended on all lines of output.
+ * @param memory The memory to be dumped.
+ * @param len The length of the memory to be dumped.
+ */
+extern void log_dumpHex(TCHAR *label, TCHAR *memory, size_t len);
+#endif
+
+/**
+ * Sets the number of milliseconds to allow logging to take before a warning is logged.
+ *  Defaults to 0 for no limit.  Possible values 0 to 3600000.
+ *
+ * @param threshold Warning threashold.
+ */
+extern void setLogWarningThreshold(int threshold);
+
 /**
  * Sets the log levels to a silence so we never output anything.
  */
@@ -151,6 +173,13 @@ extern void setSilentLogLevels();
  */
 extern void setSimpleLogLevels();
 
+/**
+ * Used for testing to set a pause into the next log entry made.
+ *
+ * @param pauseTime Number of seconds to pause, 0 pauses indefinitely.
+ */
+extern void setPauseTime(int pauseTime);
+
 /* * Logfile functions * */
 extern int isLogfileAccessed();
 
