@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2010 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2011 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -117,8 +117,8 @@ int multiByteToWideChar(const char *multiByteChars, const char *multiByteEncodin
     const TCHAR *errorTemplate;
     size_t errorTemplateLen;
     size_t iconv_value;
-    char *nativeChar;
     char *nativeCharStart;
+    char *nativeCharStartCopy;
     size_t multiByteCharsLen;
     size_t nativeCharLen;
     size_t nativeCharLenCopy;
@@ -143,7 +143,7 @@ int multiByteToWideChar(const char *multiByteChars, const char *multiByteEncodin
         if (conv_desc == (iconv_t)(-1)) {
             /* Initialization failure. */
             if (errno == EINVAL) {
-                errorTemplate = (localizeErrorMessage ? TEXT("Conversion from '%s' to '%s' is not supported.") : TEXT("Conversion from '%s' to '%s' is not supported."));
+                errorTemplate = (localizeErrorMessage ? TEXT("Conversion from '% s' to '% s' is not supported.") : TEXT("Conversion from '% s' to '% s' is not supported."));
                 errorTemplateLen = _tcslen(errorTemplate) + strlen(multiByteEncoding) + strlen(interumEncoding) + 1;
                 *outputBufferW = malloc(sizeof(TCHAR) * errorTemplateLen);
                 if (*outputBufferW) {
@@ -180,22 +180,27 @@ int multiByteToWideChar(const char *multiByteChars, const char *multiByteEncodin
         
         /* We need to figure out how many bytes we need to store the native encoded string. */
         nativeCharLen = multiByteCharsLen;
+        nativeCharStart = NULL;
         do {
             redoIConv = FALSE;
+            
+            if (nativeCharStart) {
+                free(nativeCharStart);
+            }
+            
             multiByteCharsLenStart = multiByteCharsLen;
             multiByteCharsStart = (char *)multiByteChars;
-            nativeChar = malloc(nativeCharLen);
-            if (!nativeChar) {
+            nativeCharStart = malloc(nativeCharLen);
+            if (!nativeCharStart) {
                 /* Out of memory. */
                 *outputBufferW = NULL;
                 return TRUE;
             }
-
-            nativeCharStart = nativeChar;
             
             /* Make a copy of the nativeCharLen as this call will replace it with the number of chars used. */
             nativeCharLenCopy = nativeCharLen;
-            iconv_value = iconv(conv_desc, &multiByteCharsStart, &multiByteCharsLenStart, &nativeChar, &nativeCharLenCopy);
+            nativeCharStartCopy = nativeCharStart;
+            iconv_value = iconv(conv_desc, &multiByteCharsStart, &multiByteCharsLenStart, &nativeCharStartCopy, &nativeCharLenCopy);
              /* Handle failures. */
             if (iconv_value == (size_t)-1) {
                 /* See "man 3 iconv" for an explanation. */
@@ -266,6 +271,25 @@ int multiByteToWideChar(const char *multiByteChars, const char *multiByteEncodin
 
     /* now store the result into a wchar_t */
     wideCharLen = mbstowcs(NULL, nativeCharStart, 0);
+    if (wideCharLen == -1) {
+        if (didIConv) {
+            free(nativeCharStart);
+        }
+        if (errno == EILSEQ) {
+            errorTemplate = (localizeErrorMessage ? TEXT("Invalid multibyte sequence.") : TEXT("Invalid multibyte sequence."));
+            errorTemplateLen = _tcslen(errorTemplate) + 1;
+        } else {
+            errorTemplate = (localizeErrorMessage ? TEXT("Unexpected iconv error: %d") : TEXT("Unexpected iconv error: %d"));
+            errorTemplateLen = _tcslen(errorTemplate) + 10 + 1;
+        }
+        *outputBufferW = malloc(sizeof(TCHAR) * errorTemplateLen);
+        if (*outputBufferW) {
+            _sntprintf(*outputBufferW, errorTemplateLen, errorTemplate, errno);
+        } else {
+            /* Out of memory. *outputBufferW already NULL. */
+        }
+        return TRUE;
+    }
     *outputBufferW = malloc(sizeof(wchar_t) * (wideCharLen + 1));
     if (!(*outputBufferW)) {
         /* Out of memory. *outputBufferW already NULL. */
@@ -303,27 +327,33 @@ size_t _treadlink(TCHAR* exe, TCHAR* fullPath, size_t size) {
             req = readlink(cExe, cFullPath, size);
             mbstowcs(fullPath, cFullPath, size);
             free(cFullPath);
+            free(cExe);
             return req * sizeof(TCHAR);
+        } else {
+            free(cExe);
         }
-        free(cExe);
     }
     return -1;
 }
 
+/**
+ * This Wrapper function internally does a malloc to generate the
+ *  Wide-char version of the return string.  This must be freed by the caller.
+ */
 TCHAR* _tgetcwd(TCHAR *buf, size_t size) {
     char* cBuf;
-    char *ptr;
-
-    cBuf = malloc(size);
-    if (cBuf) {
-        wcstombs(cBuf, buf, size);
-        ptr = (char *)getcwd(cBuf, size);
-        if (ptr != NULL) {
-            mbstowcs(buf, cBuf, size * sizeof(TCHAR));
+    if (buf) {
+        cBuf = malloc(size);
+        if (cBuf) {
+            if (getcwd(cBuf, size) != NULL) {
+                mbstowcs(buf, cBuf, size * sizeof(TCHAR));
+                free(cBuf);
+                return buf;
+            }
+            free(cBuf);
         }
-        free(cBuf);
-    }
-    return buf;
+    } 
+    return NULL;
 }
 
 long _tpathconf(const TCHAR *path, int name) {
@@ -348,6 +378,9 @@ long _tpathconf(const TCHAR *path, int name) {
 /**
  * Set the current locale.
  *
+ * This Wrapper function internally does a malloc to generate the
+ *  Wide-char version of the return string.  This must be freed by the caller.
+ *
  * @param category
  * @param locale The requested locale. TEXT("") for the default.
  *
@@ -368,18 +401,17 @@ TCHAR *_tsetlocale(int category, const TCHAR *locale) {
         wcstombs(cLocale, locale, req + 1);
         cReturn = setlocale(category, cLocale);
         free(cLocale);
-        if(!cReturn) {
-            return NULL;
+        
+        if (cReturn) {
+            req = mbstowcs(NULL, cReturn, 0);
+            if (req >= 0) {
+                tReturn = malloc(sizeof(TCHAR) * (req + 1));
+                if (tReturn) {
+                    mbstowcs(tReturn, cReturn, req + 1);
+                    return tReturn;
+                }
+            }
         }
-        req = mbstowcs(NULL, cReturn, 0);
-        if (req < 0) {
-            return NULL;
-        }
-        tReturn = malloc(sizeof(TCHAR) * (req + 1));
-        if (tReturn) {
-            mbstowcs(tReturn, cReturn, req + 1);
-        }
-        return tReturn;
     }
     return NULL;
 }
@@ -392,7 +424,7 @@ int _tprintf(const wchar_t *fmt,...) {
     if (wcsstr(fmt, TEXT("%s")) != NULL) {
         msg = malloc(sizeof(wchar_t) * (wcslen(fmt) + 1));
         if (msg) {
-            wcscpy(msg, fmt);
+            wcsncpy(msg, fmt, wcslen(fmt) + 1);
             for (i = 0; i < wcslen(fmt); i++){
                 if (fmt[i] == TEXT('%') && i  < wcslen(fmt) && fmt[i + 1] == TEXT('s') && (i == 0 || fmt[i - 1] != TEXT('%'))) {
                     msg[i + 1] = TEXT('S');
@@ -402,11 +434,11 @@ int _tprintf(const wchar_t *fmt,...) {
             msg[wcslen(fmt)] = TEXT('\0');
         }
         flag = TRUE;
-     } else {
+    } else {
          msg = (wchar_t*)fmt;
          flag = FALSE;
-     }
-     if (msg) {
+    }
+    if (msg) {
         va_start(args, fmt);
         i = vwprintf(msg, args);
         va_end (args);
@@ -425,7 +457,7 @@ int _ftprintf(FILE *stream, const wchar_t *fmt, ...) {
     if (wcsstr(fmt, TEXT("%s")) != NULL) {
         msg = malloc(sizeof(wchar_t) * (wcslen(fmt) + 1));
         if (msg) {
-            wcscpy(msg, fmt);
+            wcsncpy(msg, fmt, wcslen(fmt) + 1);
             for (i = 0; i < wcslen(fmt); i++){
                 if (fmt[i] == TEXT('%') && i  < wcslen(fmt) && fmt[i + 1] == TEXT('s') && (i == 0 || fmt[i - 1] != TEXT('%'))) {
                     msg[i + 1] = TEXT('S');
@@ -435,11 +467,11 @@ int _ftprintf(FILE *stream, const wchar_t *fmt, ...) {
             msg[wcslen(fmt)] = TEXT('\0');
         }
         flag = TRUE;
-     } else {
+    } else {
          msg = (wchar_t*)fmt;
          flag = FALSE;
-     }
-     if (msg) {
+    }
+    if (msg) {
         va_start(args, fmt);
         i = vfwprintf(stream, msg, args);
         va_end (args);
@@ -458,7 +490,7 @@ int _sntprintf(TCHAR *str, size_t size, const TCHAR *fmt, ...) {
     if (wcsstr(fmt, TEXT("%s")) != NULL) {
         msg = malloc(sizeof(wchar_t) * (wcslen(fmt) + 1));
         if (msg) {
-            wcscpy(msg, fmt);
+            wcsncpy(msg, fmt, wcslen(fmt) + 1);
             for (i = 0; i < wcslen(fmt); i++){
                 if (fmt[i] == TEXT('%') && i  < wcslen(fmt) && fmt[i + 1] == TEXT('s') && (i == 0 || fmt[i - 1] != TEXT('%'))) {
                     msg[i + 1] = TEXT('S');
@@ -468,11 +500,11 @@ int _sntprintf(TCHAR *str, size_t size, const TCHAR *fmt, ...) {
             msg[wcslen(fmt)] = TEXT('\0');
         }
         flag = TRUE;
-     } else {
+    } else {
          msg = (wchar_t*)fmt;
          flag = FALSE;
-     }
-     if (msg) {
+    }
+    if (msg) {
         va_start(args, fmt);
         i = vswprintf(str, size, msg, args);
         va_end (args);
@@ -549,7 +581,11 @@ void _tsyslog(int priority, const TCHAR *message) {
     }
 }
 
-TCHAR * _tgetenv ( const TCHAR * name ) {
+/**
+ * This Wrapper function internally does a malloc to generate the
+ *  Wide-char version of the return string.  This must be freed by the caller.
+ */
+TCHAR * _tgetenv( const TCHAR * name ) {
     char* cName;
     TCHAR* val;
     size_t req;
@@ -579,7 +615,7 @@ FILE* _tfopen(const wchar_t* file, const wchar_t* mode) {
     int sizeFile, sizeMode;
     char* cFile;
     char* cMode;
-    FILE *f;
+    FILE *f = NULL;
 
     sizeFile = wcstombs(NULL, (wchar_t*)file, 0) + 1;
     cFile= malloc(sizeFile);
@@ -591,11 +627,10 @@ FILE* _tfopen(const wchar_t* file, const wchar_t* mode) {
             wcstombs(cMode, (wchar_t*) mode, sizeMode);
             f = fopen(cFile, cMode);
             free(cMode);
-            return f;
         }
         free(cFile);
     }
-    return NULL;
+    return f;
 }
 
 int _tunlink(const wchar_t* address) {
@@ -613,6 +648,23 @@ int _tunlink(const wchar_t* address) {
     return -1;
 }
 
+
+int _tmkfifo(TCHAR* arg, mode_t mode) {
+    size_t size;
+    char *cStr;
+    int r; 
+
+    r = -1;
+    size = wcstombs(NULL, arg, 0) + 1;
+    cStr = malloc(size);
+    if (cStr) {
+        wcstombs(cStr, arg, size);
+        r = mkfifo(cStr, mode);
+        free(cStr);
+    }
+    return r;
+}
+
 int _tchdir(const TCHAR *path) {
     int r;
     size_t size;
@@ -640,7 +692,6 @@ int _texecvp(TCHAR* arg, TCHAR **cmd) {
     }
     size = i;
     cCmd = malloc((i + 1) * sizeof *cCmd);
-
     if (cCmd) {
         for (i = 0; i < size; i++) {
             req  = wcstombs(NULL, cmd[i], 0) + 1;
@@ -666,8 +717,9 @@ int _texecvp(TCHAR* arg, TCHAR **cmd) {
         } else {
             i = -1;
         }
-        for (; size >= 0; size--)
+        for (; size >= 0; size--) {
             free(cCmd[size]);
+        }
         free(cCmd);
         return i;
     }
@@ -706,7 +758,7 @@ int _vsntprintf(wchar_t *ws, size_t n, const wchar_t *format, va_list arg) {
 #endif
 
 int _texecve(TCHAR* arg, TCHAR **cmd, TCHAR** env) {
-    char** cCmd;
+    char **cCmd;
     char *cArg;
     char **cEnv;
     int i, sizeCmd, sizeEnv;
@@ -738,6 +790,13 @@ int _texecve(TCHAR* arg, TCHAR **cmd, TCHAR** env) {
         }
         sizeEnv = i;
         cEnv = malloc((i + 1) * sizeof *cEnv);
+        if (!cEnv) {
+            for (; sizeCmd >= 0; sizeCmd--) {
+                free(cCmd[sizeCmd]);
+            }
+            free(cCmd);
+            return -1;
+        }
         for (i = 0; i < sizeEnv; i++) {
             req = wcstombs(NULL, env[i], 0) + 1;
             cEnv[i] = malloc(req);
@@ -749,6 +808,10 @@ int _texecve(TCHAR* arg, TCHAR **cmd, TCHAR** env) {
                     free(cEnv[i]);
                 }
                 free(cEnv);
+                for (; sizeCmd >= 0; sizeCmd--) {
+                    free(cCmd[sizeCmd]);
+                }
+                free(cCmd);
                 return -1;
             }
         }
@@ -762,20 +825,20 @@ int _texecve(TCHAR* arg, TCHAR **cmd, TCHAR** env) {
         } else {
             i = -1;
         }
-        for (; sizeCmd >= 0; sizeCmd--) {
-            free(cCmd[sizeCmd]);
-        }
         for (; sizeEnv >= 0; sizeEnv--) {
             free(cEnv[sizeEnv]);
         }
         free(cEnv);
+        for (; sizeCmd >= 0; sizeCmd--) {
+            free(cCmd[sizeCmd]);
+        }
         free(cCmd);
         return i;
     }
     return -1;
 }
 
-int _topen(const TCHAR *path, int oflag, ...) {
+int _topen(const TCHAR *path, int oflag, mode_t mode) {
     char* cPath;
     int r;
     size_t size;
@@ -784,14 +847,18 @@ int _topen(const TCHAR *path, int oflag, ...) {
     cPath = malloc(size);
     if (cPath) {
         wcstombs(cPath, path, size);
-        r = open(cPath, oflag, 0);
+        r = open(cPath, oflag, mode);
         free(cPath);
         return r;
     }
     return -1;
 }
 
-int _tputenv(TCHAR *string) {
+#if defined(WRAPPER_USE_PUTENV)
+/**
+ * Normal calls to putenv do not free the string parameter, but UNICODE calls can and should.
+ */
+int _tputenv(const TCHAR *string) {
     int r;
     size_t size;
     char *cStr;
@@ -801,11 +868,54 @@ int _tputenv(TCHAR *string) {
     if (cStr) {
         wcstombs(cStr, string, size);
         r = putenv(cStr);
-      /*  free(cstr); */
+        /* Can't free cStr as it becomes part of the environment. */
+        /*  free(cstr); */
         return r;
     }
     return -1;
 }
+#else
+int _tsetenv(const TCHAR *name, const TCHAR *value, int overwrite) {
+    int r = -1;
+    size_t size;
+    char *cName;
+    char *cValue;
+
+    size = wcstombs(NULL, (wchar_t*)name, 0) + 1;
+    cName = malloc(size);
+    if (cName) {
+        wcstombs(cName, name, size);
+        
+        size = wcstombs(NULL, (wchar_t*)value, 0) + 1;
+        cValue = malloc(size);
+        if (cValue) {
+            wcstombs(cValue, value, size);
+            
+            r = setenv(cName, cValue, overwrite);
+            
+            free(cValue);
+        }
+        
+        free(cName);
+    }
+    return r;
+}
+
+void _tunsetenv(const TCHAR *name) {
+    size_t size;
+    char *cName;
+
+    size = wcstombs(NULL, (wchar_t*)name, 0) + 1;
+    cName = malloc(size);
+    if (cName) {
+        wcstombs(cName, name, size);
+        
+        unsetenv(cName);
+        
+        free(cName);
+    }
+}
+#endif
 
 int _tstat(const wchar_t* filename, struct stat *buf) {
     int size;
@@ -818,33 +928,51 @@ int _tstat(const wchar_t* filename, struct stat *buf) {
         size = stat(cFileName, buf);
         free(cFileName);
     }
-    return(size);
+    return size;
 }
 
-wchar_t* _trealpath(const wchar_t* file_name, wchar_t *resolved_name) {
+/**
+ * @param file_name The file name to be resolved.
+ * @param resolved_name A buffer large enough to hold MAX_PATH characters (plus a '\0')
+ *
+ * @return resolved_name of success, otherwise NULL.
+ */
+wchar_t* _trealpath(const wchar_t* fileName, wchar_t *resolvedName) {
     char *cFile;
 #if defined(IRIX)
-    char resolved[FILENAME_MAX];
+    char resolved[FILENAME_MAX + 1];
 #else
-    char resolved[PATH_MAX];
+    char resolved[PATH_MAX + 1];
 #endif
-    int sizeFile, sizeReturn;
+    int sizeFile;
+    int sizeReturn;
     char* returnVal;
+    
+    /* Initialize the return value. */
+    resolvedName[0] = TEXT('\0');
 
-    sizeFile = wcstombs(NULL, file_name, 0) + 1;
+    sizeFile = wcstombs(NULL, fileName, 0) + 1;
     cFile = malloc(sizeFile);
     if (cFile) {
-        wcstombs(cFile, file_name, sizeFile );
+        wcstombs(cFile, fileName, sizeFile);
         returnVal = realpath(cFile, resolved);
-        free(cFile);
-        if(returnVal == NULL) {
+        if (returnVal == NULL) {
+            free(cFile);
+            
+            /* The resolved var contains an error path.  Convert it. */
+            sizeReturn = mbstowcs(NULL, resolved, 0) + 1;
+            sizeReturn *= sizeof(TCHAR);
+            mbstowcs(resolvedName, resolved, sizeReturn);
+            
             return NULL;
         }
+        free(cFile);
+        
         sizeReturn = mbstowcs(NULL, resolved, 0) + 1;
         sizeReturn *= sizeof(TCHAR);
-        mbstowcs(resolved_name, resolved, sizeReturn);
-
-        return resolved_name;
+        mbstowcs(resolvedName, resolved, sizeReturn);
+    
+        return resolvedName;
     }
     return NULL;
 }
