@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2010 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2011 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -77,20 +77,27 @@ pid_t getsid(pid_t pid);
 #define max(x,y) (((x) > (y)) ? (x) : (y))
 #define min(x,y) (((x) < (y)) ? (x) : (y))
 
-int jvmOut = -1;
 
 /* Define a global pipe descriptor so that we don't have to keep allocating
  *  a new pipe each time a JVM is launched. */
-int pipedes[2];
-int pipeInitialized = 0;
+int pipedes[2] = {-1, -1};
+#define PIPE_READ_END 0
+#define PIPE_WRITE_END 1
 
 TCHAR wrapperClasspathSeparator = TEXT(':');
 
+int javaIOThreadSet = FALSE;
+pthread_t javaIOThreadId;
+int javaIOThreadStarted = FALSE;
+int stopJavaIOThread = FALSE;
+int javaIOThreadStopped = FALSE;
+
 int timerThreadSet = FALSE;
 pthread_t timerThreadId;
 int timerThreadStarted = FALSE;
 int stopTimerThread = FALSE;
 int timerThreadStopped = FALSE;
+
 TICKS timerTicks = WRAPPER_TICK_INITIAL;
 
 /******************************************************************************
@@ -217,7 +224,8 @@ void takeSignalAction(int sigNum, const TCHAR *sigName, int mode) {
                 (wrapperData->jState == WRAPPER_JSTATE_STOPPED) ||
                 (wrapperData->jState == WRAPPER_JSTATE_KILLING) ||
                 (wrapperData->jState == WRAPPER_JSTATE_KILL) ||
-                (wrapperData->jState == WRAPPER_JSTATE_DOWN_CHECK)) {
+                (wrapperData->jState == WRAPPER_JSTATE_DOWN_CHECK) ||
+                (wrapperData->jState == WRAPPER_JSTATE_DOWN_FLUSH)) {
 
                 /* Signaled while we were already shutting down. */
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
@@ -230,7 +238,8 @@ void takeSignalAction(int sigNum, const TCHAR *sigName, int mode) {
             } else {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
                     TEXT("%s trapped.  Shutting down."), sigName);
-                wrapperStopProcess(0);
+                /* Always force the shutdown as this is an external event. */
+                wrapperStopProcess(0, TRUE);
             }
             /* Don't actually kill the process here.  Let the application shut itself down */
 
@@ -632,6 +641,121 @@ int registerSigAction(int sigNum, void (*sigAction)(int, siginfo_t *, void *)) {
     return 0;
 }
 
+/**
+ * The main entry point for the javaio thread which is started by
+ *  initializeJavaIO().  Once started, this thread will run for the
+ *  life of the process.
+ *
+ * This thread will only be started if we are configured to use a
+ *  dedicated thread to read JVM output.
+ */
+void *javaIORunner(void *arg) {
+    sigset_t signal_mask;
+    int nextSleep;
+#ifndef VALGRIND
+    int rc;
+#endif
+
+    javaIOThreadStarted = TRUE;
+    
+    /* Immediately register this thread with the logger. */
+    logRegisterThread(WRAPPER_THREAD_JAVAIO);
+
+    /* mask signals so the javaIO doesn't get any of these. */
+    sigemptyset(&signal_mask);
+    sigaddset(&signal_mask, SIGTERM);
+    sigaddset(&signal_mask, SIGINT);
+    sigaddset(&signal_mask, SIGQUIT);
+    sigaddset(&signal_mask, SIGALRM);
+    sigaddset(&signal_mask, SIGHUP);
+    sigaddset(&signal_mask, SIGUSR1);
+#ifndef VALGRIND
+    sigaddset(&signal_mask, SIGUSR2);
+    rc = pthread_sigmask(SIG_BLOCK, &signal_mask, NULL);
+    if (rc != 0) {
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("Could not mask signals for javaIO thread."));
+    }
+#endif
+
+    if (wrapperData->isJavaIOOutputEnabled) {
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("JavaIO thread started."));
+    }
+
+    nextSleep = TRUE;
+    /* Loop until we are shutting down, but continue as long as there is more output from the JVM. */
+    while ((!stopJavaIOThread) || (!nextSleep)) {
+        if (nextSleep) {
+            /* Sleep for a hundredth of a second. */
+            wrapperSleep(10);
+        }
+        nextSleep = TRUE;
+        
+        if (wrapperData->pauseThreadJavaIO) {
+            wrapperPauseThread(wrapperData->pauseThreadJavaIO, TEXT("javaio"));
+            wrapperData->pauseThreadJavaIO = 0;
+        }
+        
+        if (wrapperReadChildOutput()) {
+            if (wrapperData->isDebugging) {
+                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_DEBUG,
+                    TEXT("Pause reading child process output to share cycles."));
+            }
+            nextSleep = FALSE;
+        }
+    }
+
+    javaIOThreadStopped = TRUE;
+    if (wrapperData->isJavaIOOutputEnabled) {
+        log_printf_queue(TRUE, WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("JavaIO thread stopped."));
+    }
+    return NULL;
+}
+
+/**
+ * Creates a process whose job is to loop and simply increment a ticks
+ *  counter.  The tick counter can then be used as a clock as an alternative
+ *  to using the system clock.
+ */
+int initializeJavaIO() {
+    int res;
+
+    if (wrapperData->isJavaIOOutputEnabled) {
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Launching JavaIO thread."));
+    }
+
+    res = pthread_create(&javaIOThreadId,
+        NULL, /* No attributes. */
+        javaIORunner,
+        NULL); /* No parameters need to be passed to the thread. */
+    if (res) {
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL,
+            TEXT("Unable to create a javaIO thread: %d, %s"), res, getLastErrorText());
+        javaIOThreadSet = TRUE;
+        return 1;
+    } else {
+        if (pthread_detach(javaIOThreadId)) {
+            javaIOThreadSet = TRUE;
+            return 1;
+        }
+        javaIOThreadSet = FALSE;
+        return 0;
+    }
+}
+
+void disposeJavaIO() {
+    stopJavaIOThread = TRUE;
+    /* Wait until the javaIO thread is actually stopped to avoid timing problems. */
+    if (javaIOThreadStarted) {
+        while (!javaIOThreadStopped) {
+#ifdef _DEBUG
+            wprintf(TEXT("Waiting for javaIO thread to stop.\n"));
+#endif
+            wrapperSleep(100);
+        }
+        pthread_kill(javaIOThreadId, SIGKILL);
+    }
+}
+
 /**
  * The main entry point for the timer thread which is started by
  *  initializeTimer().  Once started, this thread will run for the
@@ -639,11 +763,6 @@ int registerSigAction(int sigNum, void (*sigAction)(int, siginfo_t *, void *)) {
  *
  * This thread will only be started if we are configured NOT to
  *  use the system time as a base for the tick counter.
- *
- * All logging within this thread is intentionally using the Queued
- *  logging.  This is not because of lock problems, but rather because
- *  it is much faster and will be less likely to cause any delays in
- *  the looping of the timer.
  */
 void *timerRunner(void *arg) {
     TICKS sysTicks;
@@ -684,6 +803,11 @@ void *timerRunner(void *arg) {
 
     while (!stopTimerThread) {
         wrapperSleep(WRAPPER_TICK_MS);
+        
+        if (wrapperData->pauseThreadTimer) {
+            wrapperPauseThread(wrapperData->pauseThreadTimer, TEXT("timer"));
+            wrapperData->pauseThreadTimer = 0;
+        }
 
         /* Get the tick count based on the system time. */
         sysTicks = wrapperGetSystemTicks();
@@ -731,6 +855,9 @@ void *timerRunner(void *arg) {
     }
 
     timerThreadStopped = TRUE;
+    if (wrapperData->isTickOutputEnabled) {
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Timer thread stopped."));
+    }
     return NULL;
 }
 
@@ -825,6 +952,17 @@ int wrapperInitializeRun() {
             return res;
         }
     }
+    
+    if (wrapperData->useJavaIOThread) {
+        /* Create and initialize a javaIO thread. */
+        if ((res = initializeJavaIO()) != 0) {
+            return res;
+        }
+    } else {
+        javaIOThreadSet = FALSE;
+        /* Unable to set the javaIOThreadId to a null value on all platforms
+         * javaIOThreadId = 0;*/
+    }
 
     return retval;
 }
@@ -890,6 +1028,28 @@ int wrapperSleep(int ms) {
     return FALSE;
 }
 
+/**
+ * Detaches the Java process so the Wrapper will if effect forget about it.
+ */
+void wrapperDetachJava() {
+    int fd;
+    
+    wrapperSetJavaState(WRAPPER_JSTATE_DOWN_CLEAN, 0, -1);
+    
+    /* Redirect the pipes with the JVM to /dev/null so the JVM doesn't block when it writes to them. */
+    fd = _topen(TEXT("/dev/null"), O_RDWR, 0);
+    if (fd != -1) {
+        close(pipedes[STDOUT_FILENO]);
+        /*dup2(fd, pipedes[STDOUT_FILENO]); */
+        close(pipedes[STDERR_FILENO]);
+        /* dup2(fd, pipedes[STDERR_FILENO]); */
+        /*if ((fd != pipedes[STDOUT_FILENO]) && (fd != pipedes[STDERR_FILENO])) { */
+        close(fd);
+        /*}*/
+    }
+}
+
+
 /**
  * Build the java command line.
  *
@@ -976,15 +1136,11 @@ int wrapperBuildJavaCommand() {
 void wrapperExecute() {
     pid_t proc;
 
-    /* Only allocate a pipe if we have not already done so. */
-    if (!pipeInitialized) {
-        /* Create the pipe. */
-        if (pipe(pipedes) < 0) {
-            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
-                       TEXT("Could not init pipe: %s"), getLastErrorText());
-            return;
-        }
-        pipeInitialized = TRUE;
+    /* Create the pipe. */
+    if (pipe(pipedes) < 0) {
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
+                   TEXT("Could not init pipe: %s"), getLastErrorText());
+        return;
     }
 
     /* Make sure the log file is closed before the Java process is created.  Failure to do
@@ -1028,18 +1184,24 @@ void wrapperExecute() {
              * TODO: Figure out a way to fix this.  Maybe using shared memory? */
 
             /* Send output to the pipe by dupicating the pipe fd and setting the copy as the stdout fd. */
-            if (dup2(pipedes[STDOUT_FILENO], STDOUT_FILENO) < 0) {
+            if (dup2(pipedes[PIPE_WRITE_END], STDOUT_FILENO) < 0) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
                     TEXT("%sUnable to set JVM's stdout: %s"), LOG_FORK_MARKER, getLastErrorText());
                 return;
             }
 
             /* Send errors to the pipe by dupicating the pipe fd and setting the copy as the stderr fd. */
-            if (dup2(pipedes[STDOUT_FILENO], STDERR_FILENO) < 0) {
+            if (dup2(pipedes[PIPE_WRITE_END], STDERR_FILENO) < 0) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
                     TEXT("%sUnable to set JVM's stderr: %s"), LOG_FORK_MARKER, getLastErrorText());
                 return;
             }
+            
+            /* Close both ends of the pipe as we have already duplicated the Write end for our purposes. */
+            close(pipedes[PIPE_READ_END]);
+            pipedes[PIPE_READ_END] = -1;
+            close(pipedes[PIPE_WRITE_END]);
+            pipedes[PIPE_WRITE_END] = -1;
 
             /* The pipedes array is global so do not close the pipes. */
             /* Child process: execute the JVM. */
@@ -1075,7 +1237,10 @@ void wrapperExecute() {
         } else {
             /* We are the parent side. */
             wrapperData->javaPID = proc;
-            jvmOut = pipedes[STDIN_FILENO];
+            
+            /* Close the write end as it is not used. */
+            close(pipedes[PIPE_WRITE_END]);
+            pipedes[PIPE_WRITE_END] = -1;
 
             /* Restore the auto close flag. */
             setLogfileAutoClose(wrapperData->logfileInactivityTimeout <= 0);
@@ -1084,12 +1249,12 @@ void wrapperExecute() {
 
             /* Mark our side of the pipe so that it won't block
              * and will close on exec, so new children won't see it. */
-            if (fcntl(jvmOut, F_SETFL, O_NONBLOCK) < 0) {
+            if (fcntl(pipedes[PIPE_READ_END], F_SETFL, O_NONBLOCK) < 0) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
                     TEXT("Failed to set JVM output handle to non blocking mode: %s (%d)"),
                     getLastErrorText(), errno);
             }
-            if (fcntl(jvmOut, F_SETFD, FD_CLOEXEC) < 0) {
+            if (fcntl(pipedes[PIPE_READ_END], F_SETFD, FD_CLOEXEC) < 0) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
                     TEXT("Failed to set JVM output handle to close on JVM exit: %s (%d)"),
                     getLastErrorText(), errno);
@@ -1242,9 +1407,10 @@ int wrapperGetProcessStatus(TICKS nowTicks, int sigChild) {
         res = WRAPPER_PROCESS_UP;
     } else if (retval < 0) {
         if (errno == ECHILD) {
-            if (wrapperData->jState == WRAPPER_JSTATE_DOWN_CHECK ||
-                wrapperData->jState == WRAPPER_JSTATE_DOWN_CLEAN ||
-                wrapperData->jState == WRAPPER_JSTATE_STOPPED) {
+            if ((wrapperData->jState == WRAPPER_JSTATE_DOWN_CHECK) ||
+                (wrapperData->jState == WRAPPER_JSTATE_DOWN_FLUSH) ||
+                (wrapperData->jState == WRAPPER_JSTATE_DOWN_CLEAN) ||
+                (wrapperData->jState == WRAPPER_JSTATE_STOPPED)) {
                 res = WRAPPER_PROCESS_DOWN;
                 wrapperJVMProcessExited(nowTicks, 0);
                 return res;
@@ -1311,7 +1477,7 @@ void wrapperReportStatus(int useLoggerQueue, int status, int errorCode, int wait
  * Returns TRUE if there were any problems, FALSE otherwise.
  */
 int wrapperReadChildOutputBlock(char *blockBuffer, int blockSize, int *readCount) {
-    if (jvmOut == -1) {
+    if (pipedes[PIPE_READ_END] == -1) {
         /* The child is not up. */
         *readCount = 0;
         return FALSE;
@@ -1322,15 +1488,15 @@ int wrapperReadChildOutputBlock(char *blockBuffer, int blockSize, int *readCount
      *  http://www.freebsd.org/cgi/query-pr.cgi?pr=kern/64313
      *
      * When linked with the pthreads library the O_NONBLOCK flag is being reset
-     *  on the jvmOut handle.  Not sure yet of the exact event that is causing
+     *  on the pipedes[PIPE_READ_END] handle.  Not sure yet of the exact event that is causing
      *  this, but once it happens reads will start to block even though calls
-     *  to fcntl(jvmOut, F_GETFL) say that the O_NONBLOCK flag is set.
-     * Calling fcntl(jvmOut, F_SETFL, O_NONBLOCK) again will set the flag back
+     *  to fcntl(pipedes[PIPE_READ_END], F_GETFL) say that the O_NONBLOCK flag is set.
+     * Calling fcntl(pipedes[PIPE_READ_END], F_SETFL, O_NONBLOCK) again will set the flag back
      *  again and cause it to start working correctly.  This may only need to
      *  be done once, however, because F_GETFL does not return the accurate
      *  state there is no reliable way to check.  Be safe and always set the
      *  flag. */
-    if (fcntl(jvmOut, F_SETFL, O_NONBLOCK) < 0) {
+    if (fcntl(pipedes[PIPE_READ_END], F_SETFL, O_NONBLOCK) < 0) {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT(
             "Failed to set JVM output handle to non blocking mode to read child process output: %s (%d)"),
             getLastErrorText(), errno);
@@ -1339,9 +1505,8 @@ int wrapperReadChildOutputBlock(char *blockBuffer, int blockSize, int *readCount
 #endif
 
     /* Fill read buffer. */
-    *readCount = read(jvmOut, blockBuffer, blockSize);
-
-    if (*readCount <= 0) {
+    *readCount = read(pipedes[PIPE_READ_END], blockBuffer, blockSize);
+    if (*readCount < 0) {
         /* No more bytes available, return for now.  But make sure that this was not an error. */
         if (errno == EAGAIN) {
             /* Normal, the call would have blocked as there is no data available. */
@@ -1351,6 +1516,10 @@ int wrapperReadChildOutputBlock(char *blockBuffer, int blockSize, int *readCount
                 getLastErrorText(), errno);
             return TRUE;
         }
+    } else if (*readCount == 0) {
+        /* We reached the EOF.  This means that the other end of the pipe was closed. */
+        close(pipedes[PIPE_READ_END]);
+        pipedes[PIPE_READ_END] = -1;
     }
 
     return FALSE;
@@ -1358,9 +1527,6 @@ int wrapperReadChildOutputBlock(char *blockBuffer, int blockSize, int *readCount
 
 /**
  * Transform a program into a daemon.
- * Inspired by code from GNU monit, which in turn, was
- * inspired by code from Stephen A. Rago's book,
- * Unix System V Network Programming.
  *
  * The idea is to first fork, then make the child a session leader,
  * and then fork again, so that it, (the session group leader), can
@@ -1573,6 +1739,10 @@ int main(int argc, char **argv) {
         return 1; /* For compiler. */
     }
     wrapperLoadHostName();
+    if (!strcmpIgnoreCase(wrapperData->argCommand, TEXT("-translate"))) {
+        /* We want to disable all log output when a translation request is made. */
+        setSilentLogLevels();
+    }
     /* At this point, we have a command, confFile, and possibly additional arguments. */
     if (!strcmpIgnoreCase(wrapperData->argCommand, TEXT("?")) || !strcmpIgnoreCase(wrapperData->argCommand, TEXT("-help"))) {
         /* User asked for the usage. */
@@ -1610,8 +1780,13 @@ int main(int argc, char **argv) {
 
     /* Set the default umask of the Wrapper process. */
     umask(wrapperData->umask);
-
-    if (!strcmpIgnoreCase(wrapperData->argCommand, TEXT("c")) || !strcmpIgnoreCase(wrapperData->argCommand, TEXT("-console"))) {
+    if (!strcmpIgnoreCase(wrapperData->argCommand, TEXT("-translate"))) {
+        setSimpleLogLevels();
+        /* Print out the string so the caller sees it as its translated output. */
+        _tprintf(TEXT("%s"), argv[2]);
+        appExit(0, argc, argv);
+        return 0; /* For compiler. */
+    } else if (!strcmpIgnoreCase(wrapperData->argCommand, TEXT("c")) || !strcmpIgnoreCase(wrapperData->argCommand, TEXT("-console"))) {
         /* Run as a console application */
 
         /* fork to a Daemonized process if configured to do so. */
